###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for WaterTAP database wrapper
"""
import pytest
import os

from watertap.core.wt_database import Database


@pytest.mark.unit
def test_default_path():
    db = Database()

    assert os.path.normpath(db._dbpath) == os.path.normpath(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..", "..", "data", "techno_economic"))


@pytest.mark.unit
def test_invalid_path():
    with pytest.raises(OSError,
                       match="Could not find requested path foo. Please "
                       "check that this path exists."):
        Database(dbpath="foo")


@pytest.mark.unit
def test_custom_path():
    # Pick a path we know will exist, even if it isn't a data folder
    db = Database(dbpath=os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..", "..", "core"))

    assert db._dbpath == os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..", "..", "core")


class TestDatabase():
    @pytest.fixture(scope="class")
    def db(self):
        return Database()

    @pytest.mark.unit
    def test_component_list(self, db):
        # First, check that _component_list exists and is None
        assert db._component_list is None

        # Then check that component list retrieved is a dict
        assert isinstance(db.component_list, dict)

    @pytest.mark.unit
    def test_get_technology(self, db):
        assert db._cached_files == {}

        data = db._get_technology("nanofiltration")

        assert "nanofiltration" in db._cached_files
        assert db._cached_files["nanofiltration"] is data

        # Check for a few expected keys to check what we got back looks right
        assert "default" in data
        assert "removal_frac_mass_solute" in data["default"]
        assert "energy_electric_flow_vol_inlet" in data["default"]
        assert "recovery_frac_mass_H2O" in data["default"]

    @pytest.mark.unit
    def test_get_technology_invalid(self, db):
        with pytest.raises(KeyError,
                           match="Could not find entry for foo in database."):
            db._get_technology("foo")

        assert len(db._cached_files) == 1
        assert "foo" not in db._cached_files

    @pytest.mark.unit
    def test_get_unit_operation_parameters_default(self, db):
        data = db.get_unit_operation_parameters("nanofiltration")

        assert data == db._cached_files["nanofiltration"]["default"]

        # Check for a few expected keys to check what we got back looks right
        assert "removal_frac_mass_solute" in data
        assert "energy_electric_flow_vol_inlet" in data
        assert "recovery_frac_mass_H2O" in data

    @pytest.mark.unit
    def test_get_unit_operation_parameters_invalid_subtype(self, db):
        with pytest.raises(KeyError,
                           match="Received unrecognised subtype foo for "
                           "technology nanofiltration."):
            db.get_unit_operation_parameters("nanofiltration", subtype="foo")

    @pytest.mark.unit
    def test_get_unit_operation_parameters_single_subtype(self, db):
        # First, insert some data for a subtype into nanofiltration entry
        db._cached_files["nanofiltration"]["subtype1"] = {
            "recovery_frac_mass_H2O": "overloaded",
            "new_param": True}

        # Load data for subtype
        data = db.get_unit_operation_parameters(
            "nanofiltration", subtype="subtype1")

        # Check data
        for k, v in data.items():
            if k == "recovery_frac_mass_H2O":
                assert v == "overloaded"
            elif k == "new_param":
                assert v is True
            else:
                # All other entries should match defaults
                assert v == db._cached_files["nanofiltration"]["default"][k]

    @pytest.mark.unit
    def test_get_unit_operation_parameters_invalid_list_of_subtype(self, db):
        with pytest.raises(KeyError,
                           match="Received unrecognised subtype foo for "
                           "technology nanofiltration."):
            db.get_unit_operation_parameters(
                "nanofiltration", subtype=["subtype1", "foo"])

    @pytest.mark.unit
    def test_get_unit_operation_parameters_multi_subtype(self, db):
        # First, insert some data for a 2nd subtype into nanofiltration entry
        db._cached_files["nanofiltration"]["subtype2"] = {
            "recovery_frac_mass_H2O": "overloaded_again",
            "new_param_2": False}

        # Load data for subtype
        data = db.get_unit_operation_parameters(
            "nanofiltration", subtype="subtype2")

        # Check data
        for k, v in data.items():
            if k == "recovery_frac_mass_H2O":
                assert v == "overloaded_again"
            elif k == "new_param":
                assert v is True
            elif k == "new_param_2":
                assert v is False
            else:
                # All other entries should match defaults
                assert v == db._cached_files["nanofiltration"]["default"][k]

    @pytest.mark.unit
    def test_get_unit_operation_parameters_subtype_argument(self, db):
        with pytest.raises(TypeError,
                           match="Unexpected type for subtype 12: must be "
                           "string or list like."):
            db.get_unit_operation_parameters(
                "nanofiltration", subtype=12)

    @pytest.mark.unit
    def test_get_solute_set_default(self, db):
        comp_set = db.get_solute_set()

        assert comp_set == [
            "boron", "bromide", "calcium", "chloride", "magnesium",
            "potassium", "sodium", "strontium", "sulfate", "tds", "tss"]

    @pytest.mark.unit
    def test_get_solute_set_specified(self, db):
        comp_set = db.get_solute_set("seawater")

        assert comp_set == [
            "boron", "bromide", "calcium", "chloride", "magnesium",
            "potassium", "sodium", "strontium", "sulfate", "tds", "tss"]

    @pytest.mark.unit
    def test_get_solute_set_no_default(self, db):
        # First, delete default entry from database
        del db._cached_files["water_sources"]["default"]

        with pytest.raises(KeyError,
                           match="Database has not defined a default water "
                           "source and none was provided."):
            db.get_solute_set()

    @pytest.mark.unit
    def test_flush_cache(self, db):
        db.flush_cache()

        assert db._cached_files == {}



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order DISO unit model
"""
import pytest
from io import StringIO

from idaes.core import declare_process_block_class, FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver
import idaes.core.util.scaling as iscale
from pyomo.environ import (check_optimal_termination,
                           ConcreteModel,
                           Constraint,
                           value,
                           Var)
from pyomo.network import Port
from pyomo.util.check_units import assert_units_consistent


from watertap.core import WaterParameterBlock, WaterStateBlock, ZeroOrderBaseData
from watertap.core.zero_order_diso import (
    build_diso, initialize_diso, calculate_scaling_factors_diso, _get_Q_diso)

solver = get_solver()


@declare_process_block_class("DerivedDISO")
class DerivedDISOData(ZeroOrderBaseData):
    def build(self):
        super().build()

        build_diso(self)


class TestDISO:
    @pytest.fixture(scope="module")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.water_props = WaterParameterBlock(
            default={"solute_list": ["A", "B", "C"]})

        m.fs.unit = DerivedDISO(
            default={"property_package": m.fs.water_props})

        m.fs.unit.inlet1.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet1.flow_mass_comp[0, "A"].fix(10)
        m.fs.unit.inlet1.flow_mass_comp[0, "B"].fix(20)
        m.fs.unit.inlet1.flow_mass_comp[0, "C"].fix(30)

        m.fs.unit.inlet2.flow_mass_comp[0, "H2O"].fix(1000)
        m.fs.unit.inlet2.flow_mass_comp[0, "A"].fix(10)
        m.fs.unit.inlet2.flow_mass_comp[0, "B"].fix(20)
        m.fs.unit.inlet2.flow_mass_comp[0, "C"].fix(30)

        m.fs.unit.recovery_frac_mass_H2O.fix(0)
        m.fs.unit.removal_frac_mass_solute[0, "A"].fix(0.1)
        m.fs.unit.removal_frac_mass_solute[0, "B"].fix(0.2)
        m.fs.unit.removal_frac_mass_solute[0, "C"].fix(0.3)

        return m

    @pytest.mark.unit
    def test_private_attributes(self, model):
        assert model.fs.unit._tech_type is None
        assert model.fs.unit._has_recovery_removal is True
        assert model.fs.unit._fixed_perf_vars == []
        assert model.fs.unit._initialize is initialize_diso
        assert model.fs.unit._scaling is calculate_scaling_factors_diso
        assert model.fs.unit._get_Q is _get_Q_diso
        assert model.fs.unit._stream_table_dict == {
            "Inlet 1": model.fs.unit.inlet1,
            "Inlet 2": model.fs.unit.inlet2,
            "Treated": model.fs.unit.treated}
        assert model.fs.unit._perf_var_dict == {
            "Water Recovery": model.fs.unit.recovery_frac_mass_H2O,
            "Solute Removal": model.fs.unit.removal_frac_mass_solute}

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.fs.unit.properties_in1, WaterStateBlock)
        assert isinstance(model.fs.unit.properties_in2, WaterStateBlock)
        assert isinstance(model.fs.unit.properties_treated, WaterStateBlock)

        assert isinstance(model.fs.unit.inlet1, Port)
        assert isinstance(model.fs.unit.inlet2, Port)
        assert isinstance(model.fs.unit.treated, Port)

        assert isinstance(model.fs.unit.recovery_frac_mass_H2O, Var)
        assert isinstance(model.fs.unit.removal_frac_mass_solute, Var)
        assert len(model.fs.unit.removal_frac_mass_solute) == 3

        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert len(model.fs.unit.water_recovery_equation) == 1
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)
        assert len(model.fs.unit.solute_treated_equation) == 3

    @pytest.mark.unit
    def test_degrees_of_freedom(self, model):
        from idaes.core.util.model_statistics import unfixed_variables_set
        [print(i) for i in unfixed_variables_set(model)]
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model)

    @pytest.mark.component
    def test_scaling(self, model):
        iscale.calculate_scaling_factors(model)

        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.water_recovery_equation[0]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_treated_equation[0, "A"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_treated_equation[0, "B"]) == 1e5
        assert iscale.get_constraint_transform_applied_scaling_factor(
            model.fs.unit.solute_treated_equation[0, "C"]) == 1e5

    @pytest.mark.component
    def test_initialization(self, model):
        initialization_tester(model)

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(10e-14, rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "H2O"]))

        assert (pytest.approx(18.00, rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "A"]))
        assert (pytest.approx(32.00, rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "B"]))
        assert (pytest.approx(42.00, rel=1e-5) ==
                value(model.fs.unit.treated.flow_mass_comp[0, "C"]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()
        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                : Value   : Fixed : Bounds
    Solute Removal [A] : 0.10000 :  True : (0, None)
    Solute Removal [B] : 0.20000 :  True : (0, None)
    Solute Removal [C] : 0.30000 :  True : (0, None)
        Water Recovery :  0.0000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                            Inlet 1  Inlet 2   Treated 
    Volumetric Flowrate     1.0600   1.0600    0.092000
    Mass Concentration H2O  943.40   943.40  1.0870e-12
    Mass Concentration A    9.4340   9.4340      195.65
    Mass Concentration B    18.868   18.868      347.83
    Mass Concentration C    28.302   28.302      456.52
====================================================================================
"""

        assert output == stream.getvalue()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order well field model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    ConcreteModel, Constraint, Param, Block, value, Var, assert_optimal_termination)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import DeepWellInjectionZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()

class TestDeepWellInjectionZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["sulfur", "toc", "tss"]})

        m.fs.unit = DeepWellInjectionZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(1e-5)
        m.fs.unit.inlet.flow_mass_comp[0, "sulfur"].fix(10)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(20)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(30)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == 'deep_well_injection'
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.pipe_distance, Var)
        assert isinstance(model.fs.unit.pipe_diameter, Var)


    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("deep_well_injection")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.pipe_distance[0].fixed
        assert model.fs.unit.pipe_distance[0].value == data[
            "pipe_distance"]["value"]
        assert model.fs.unit.pipe_diameter[0].fixed
        assert model.fs.unit.pipe_diameter[0].value == data[
            "pipe_diameter"]["value"]
        assert model.fs.unit.flow_basis[0].fixed
        assert model.fs.unit.flow_basis[0].value == data[
            "flow_basis"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        for t, j in model.fs.unit.inlet.flow_mass_comp:
            assert (pytest.approx(value(
                model.fs.unit.inlet.flow_mass_comp[t, j]), rel=1e-5) ==
                    value(model.fs.unit.outlet.flow_mass_comp[t, j]))

        assert (pytest.approx(22.109214, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.outlet.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                    : Value  : Fixed : Bounds
        Electricity Demand : 22.109 : False : (0, None)
    Pipe Diameter (inches) : 8.0000 :  True : (None, None)
     Pipe Distance (miles) : 0.0000 :  True : (None, None)

------------------------------------------------------------------------------------
    Stream Table
                                 Inlet     Outlet  
    Volumetric Flowrate         0.060000   0.060000
    Mass Concentration H2O    0.00016667 0.00016667
    Mass Concentration sulfur     166.67     166.67
    Mass Concentration toc        333.33     333.33
    Mass Concentration tss        500.00     500.00
====================================================================================
"""

        assert output in stream.getvalue()

def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tss"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = DeepWellInjectionZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(1e-5)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(10)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(20)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(30)
    m.fs.unit1.load_parameters_from_database()
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.deep_well_injection, Block)
    assert isinstance(m.fs.costing.deep_well_injection.well_pump_cost,
                      Var)
    assert isinstance(m.fs.costing.deep_well_injection.pipe_cost_basis,
                      Var)
    assert isinstance(m.fs.costing.deep_well_injection.flow_exponent,
                      Var)
    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a storage tank unit.
"""

from pyomo.environ import units as pyunits, Var
from pyomo.common.config import ConfigValue, In
from idaes.core import declare_process_block_class
from watertap.core import build_pt, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("StorageTankZO")
class StorageTankZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a storage tank unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        build_pt(self)

        # self._has_recovery_removal = False

        self._tech_type = "storage_tank"

        self.storage_time = Var(self.flowsheet().time,
                                units=pyunits.hours,
                                doc="Storage time needed")

        self.surge_capacity = Var(self.flowsheet().time,
                                units=pyunits.dimensionless,
                                doc="Additional capacity needed for surge flow")

        self._fixed_perf_vars.append(self.storage_time)
        self._fixed_perf_vars.append(self.surge_capacity)

        self.tank_volume = Var(self.flowsheet().time,
                                    units=pyunits.m**3,
                                    doc="Storage tank volume")

        @self.Constraint(self.flowsheet().time,
                    doc='Tank volume constraint')
        def tank_volume_constraint(b, t):
            return b.tank_volume[t] == pyunits.convert(b.properties[t].flow_vol, to_units=pyunits.m**3/pyunits.hr) * b.storage_time[t] * (1 + b.surge_capacity[t])

        self._perf_var_dict["Storage Time (hr)"] = self.storage_time
        self._perf_var_dict["Surge Capacity (%)"] = self.surge_capacity
        self._perf_var_dict["Tank Volume (m3)"] = self.tank_volume
    


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order Ion exchange model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    Block, check_optimal_termination, ConcreteModel, Constraint, value, Var, Param)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import IonExchangeZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestIonExchangeZO_w_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tds", "foo"]})

        m.fs.unit = IonExchangeZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "tds"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(4)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

        assert isinstance(model.fs.unit.NaCl_dose, Var)
        assert isinstance(model.fs.unit.NaCl_flowrate, Var)
        assert isinstance(model.fs.unit.NaCl_constraint, Constraint)

        assert isinstance(model.fs.unit.resin_replacement, Var)
        assert isinstance(model.fs.unit.resin_demand, Var)
        assert isinstance(model.fs.unit.resin_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("ion_exchange")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(10.00410, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(9.99590e-3, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tds"]))
        assert (pytest.approx(0.39984, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(2899.6, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))
        assert (model.fs.unit.properties_in[0].flow_mass_comp["H2O"].value ==
                model.fs.unit.properties_treated[0].flow_mass_comp["H2O"].value)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-5 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                  : Value    : Fixed : Bounds
      Electricity Demand :   2899.6 : False : (0, None)
           NaCl Addition : 0.061031 : False : (0, None)
            Resin Demand : 0.013630 : False : (0, None)
    Solute Removal [foo] :   0.0000 :  True : (0, None)
    Solute Removal [tds] :  0.90000 :  True : (0, None)
          Water Recovery :   1.0000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet    Treated  Byproduct
    Volumetric Flowrate      10.005    10.004 0.00090000
    Mass Concentration H2O   999.50    999.59 1.1112e-07
    Mass Concentration tds 0.099950 0.0099959     1000.0
    Mass Concentration foo  0.39980   0.39984 1.1159e-07
====================================================================================
"""
        assert output in stream.getvalue()


class TestIonExchangeZO_clinoptilolite:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["ammonium_as_nitrogen", "foo"]})

        m.fs.unit = IonExchangeZO(default={
            "property_package": m.fs.params,
            "database": m.db,
            "process_subtype": "clinoptilolite"})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "foo"].fix(4)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.lift_height, Param)
        assert isinstance(model.fs.unit.eta_pump, Param)
        assert isinstance(model.fs.unit.eta_motor, Param)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.water_recovery_equation, Constraint)
        assert isinstance(model.fs.unit.solute_treated_equation, Constraint)

        assert isinstance(model.fs.unit.NaCl_dose, Var)
        assert isinstance(model.fs.unit.NaCl_flowrate, Var)
        assert isinstance(model.fs.unit.NaCl_constraint, Constraint)

        assert isinstance(model.fs.unit.resin_replacement, Var)
        assert isinstance(model.fs.unit.resin_demand, Var)
        assert isinstance(model.fs.unit.resin_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("ion_exchange",
                                                      subtype=model.fs.unit.config.process_subtype)

        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j == "foo":
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(9.734, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(1.027336e-4, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["ammonium_as_nitrogen"]))
        assert (pytest.approx(0.41093, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["foo"]))
        assert (pytest.approx(2899.6, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))


    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-5 >= abs(value(
                model.fs.unit.inlet.flow_mass_comp[0, j] -
                model.fs.unit.treated.flow_mass_comp[0, j] -
                model.fs.unit.byproduct.flow_mass_comp[0, j]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                         : Value      : Fixed : Bounds
                             Electricity Demand :     2899.6 : False : (0, None)
    Final mass flow of clay and nitrogen (kg/s) :     24.975 : False : (None, None)
                                  NaCl Addition : 1.0000e-08 : False : (0, None)
            Nitrogen-Clay Mixture Ratio (kg/kg) :   0.040000 :  True : (None, None)
                                   Resin Demand : 1.0000e-08 : False : (0, None)
          Solute Removal [ammonium_as_nitrogen] :    0.99900 :  True : (0, None)
                           Solute Removal [foo] :     0.0000 :  True : (0, None)
                                 Water Recovery :    0.97300 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                                              Inlet    Treated   Byproduct
    Volumetric Flowrate                       10.005     9.7340    0.27100
    Mass Concentration H2O                    999.50     999.59     996.31
    Mass Concentration ammonium_as_nitrogen 0.099950 0.00010273     3.6864
    Mass Concentration foo                   0.39980    0.41093 3.6901e-08
====================================================================================
"""
        assert output in stream.getvalue()


db = Database()
params = db._get_technology("ion_exchange")


class TestIXZOsubtype:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["tds"]})

        m.fs.unit = IonExchangeZO(default={
            "property_package": m.fs.params,
            "database": db})

        return m

    @pytest.mark.parametrize("subtype", [k for k in params.keys()])
    @pytest.mark.component
    def test_load_parameters(self, model, subtype):
        model.fs.unit.config.process_subtype = subtype
        data = db.get_unit_operation_parameters("ion_exchange", subtype=subtype)

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            if j not in data["removal_frac_mass_solute"].keys():
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]


@pytest.mark.parametrize("subtype", [k for k in params.keys()])
@pytest.mark.component
def test_costing(subtype):
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["sulfur", "toc", "tds","ammonium_as_nitrogen"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = IonExchangeZO(default={
        "property_package": m.fs.params,
        "database": m.db,
        "process_subtype": subtype})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tds"].fix(3)
    m.fs.unit1.inlet.flow_mass_comp[0, "ammonium_as_nitrogen"].fix(1)

    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.ion_exchange, Block)
    assert isinstance(m.fs.costing.ion_exchange.capital_a_parameter,
                      Var)
    assert isinstance(m.fs.costing.ion_exchange.capital_b_parameter,
                      Var)
    assert isinstance(m.fs.costing.ion_exchange.capital_c_parameter,
                      Var)
    assert isinstance(m.fs.costing.ion_exchange.capital_d_parameter,
                      Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]

    assert m.fs.unit1.NaCl_flowrate[0] in \
        m.fs.costing._registered_flows["sodium_chloride"]
    assert m.fs.unit1.resin_demand[0] in \
        m.fs.costing._registered_flows["ion_exchange_resin"]

@pytest.mark.unit
def test_clinoptilolite_no_ammonium_in_solute_list_error():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["foo"]})

    with pytest.raises(KeyError,
                       match="ammonium_as_nitrogen should be defined in "
                             "solute_list for this subtype."):
        m.fs.unit = IonExchangeZO(default={
            "property_package": m.fs.params,
            "database": db,
            "process_subtype": "clinoptilolite"})


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a chemical addition unit
operation.
"""
import pyomo.environ as pyo

from idaes.core import declare_process_block_class
from idaes.core.util.exceptions import ConfigurationError

from watertap.core import build_pt, pump_electricity, ZeroOrderBaseData

# Some more inforation about this module
__author__ = "Andrew Lee"


@declare_process_block_class("ChemicalAdditionZO")
class ChemicalAdditionZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a chemical addition unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "chemical_addition"

        if self.config.process_subtype is None:
            raise ConfigurationError(
                f"{self.name} - zero-order chemical addition operations "
                "require the process_subtype configuration argument to be set")

        build_pt(self)

        self.chemical_dosage = pyo.Var(self.flowsheet().time,
                                       units=pyo.units.mg/pyo.units.L,
                                       bounds=(0, None),
                                       doc="Dosing rate of chemical")

        self.solution_density = pyo.Var(
            bounds=(0, None),
            units=pyo.units.kg/pyo.units.m**3,
            doc="Mass density of chemical solution")
        self.ratio_in_solution = pyo.Var(
            bounds=(0, 1),
            units=pyo.units.dimensionless,
            doc="Mass fraction of chemical in solution")

        self.chemical_flow_vol = pyo.Var(
            self.flowsheet().time,
            units=pyo.units.m**3/pyo.units.s,
            bounds=(0, None),
            doc="Volumetric flow rate of chemical solution")

        self._fixed_perf_vars.append(self.chemical_dosage)
        self._fixed_perf_vars.append(self.solution_density)
        self._fixed_perf_vars.append(self.ratio_in_solution)

        self._perf_var_dict["Chemical Dosage"] = self.chemical_dosage
        self._perf_var_dict["Chemical Flow"] = self.chemical_flow_vol

        def rule_chem_flow(blk, t):
            return (blk.chemical_flow_vol[t] ==
                    pyo.units.convert(
                        blk.chemical_dosage[t]*blk.properties[t].flow_vol /
                        (blk.solution_density*blk.ratio_in_solution),
                        to_units=pyo.units.m**3/pyo.units.s))
        self.chemical_flow_constraint = pyo.Constraint(
            self.flowsheet().time, rule=rule_chem_flow)

        pump_electricity(self, self.chemical_flow_vol)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from copy import deepcopy
from enum import Enum, auto
# Import Pyomo libraries
from pyomo.environ import (Block,
                           Set,
                           Var,
                           Param,
                           Suffix,
                           NonNegativeReals,
                           Reals,
                           Reference,
                           units as pyunits,
                           log,
                           value,
                           Expr_if,
                           Constraint,
                           exp,
                           check_optimal_termination,
                           )
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.common.collections import ComponentSet
# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class,
                        MaterialBalanceType,
                        MomentumBalanceType,
                        UnitModelBlockData,
                        useDefault,
                        MaterialFlowBasis)
from idaes.core.util import get_solver
from idaes.core.util.math import smooth_min
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.scaling as iscale
from idaes.core.util.constants import Constants
from watertap.core.util.initialization import check_solve, check_dof

import idaes.logger as idaeslog


_log = idaeslog.getLogger(__name__)

#TODO: for refinement and validation phase:
# - Refine initialization routine
# - Refine initial variable values
# - Refine scaling
# - Add/refine tests
# -Add constraints for computing pressure drop in spiral wound membrane

class MassTransferCoefficient(Enum):
    fixed = auto()                   # mass transfer coefficient is a user specified value
    spiral_wound = auto()            # mass transfer coefficient is calculated using spiral wound correlation

@declare_process_block_class("NanofiltrationDSPMDE0D")
class NanofiltrationData(UnitModelBlockData):
    """
    Nanofiltration model based on Donnan Steric Pore Model with Dielectric Exclusion (DSPM-DE).

    Assumptions
        - Membrane electric potential at membrane interface is taken as reference (i.e., equal to 0)

    References:
        Geraldes and Alves, 2008 (https://doi.org/10.1016/j.memsci.2008.04.054)
        Roy et al., 2015 (http://dx.doi.org/10.1016/j.memsci.2015.06.030)
        Labban et al., 2017 (http://dx.doi.org/10.1016/j.memsci.2016.08.062)
    """
    CONFIG = ConfigBlock()

    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
    **default** = False. NF units do not support dynamic
    behavior."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. NF units do not have defined volume, thus
    this must be False."""))
    CONFIG.declare("material_balance_type", ConfigValue(
        default=MaterialBalanceType.useDefault,
        domain=In(MaterialBalanceType),
        description="Material balance construction flag",
        doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}"""))
    CONFIG.declare("momentum_balance_type", ConfigValue(
        default=MomentumBalanceType.pressureTotal,
        domain=In(MomentumBalanceType),
        description="Momentum balance construction flag",
        doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))
    CONFIG.declare("has_pressure_change", ConfigValue(
        default=False,
        domain=In([True, False]),
        description="Pressure change term construction flag",
        doc="""Indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}"""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}"""))
    CONFIG.declare("mass_transfer_coefficient", ConfigValue(
        default=MassTransferCoefficient.fixed,
        domain=In(MassTransferCoefficient),
        description="Mass transfer coefficient in feed channel",
        doc="""
            Options to account for mass transfer coefficient.

            **default** - ``MassTransferCoefficient.fixed``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``MassTransferCoefficient.fixed``", "Specify an estimated value for the mass transfer coefficient in the feed channel"
            "``MassTransferCoefficient.spiral_wound``", "Allow model to perform calculation of mass transfer coefficient based on 
            spiral wound module correlation"
        """))

    def _process_config(self):
        if len(self.config.property_package.solvent_set) > 1:
            raise ConfigurationError("NF model only supports one solvent component,"
                                     "the provided property package has specified {} solvent components"
                                     .format(len(self.config.property_package.solvent_set)))

        if len(self.config.property_package.solvent_set) == 0:
            raise ConfigurationError("The NF model was expecting a solvent and did not receive it.")

        if (hasattr(self.config.property_package,'ion_set') and len(self.config.property_package.ion_set) == 0) \
                or (hasattr(self.config.property_package,'solute_set') and len(self.config.property_package.solute_set) == 0):
            raise ConfigurationError("This NF model was expecting ions/solutes and did not receive any.")

    def build(self):
        super().build()

        self._process_config()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # For stateblock-specific scaling in calculate_scaling_factors
        self._sb_scaled_properties = ComponentSet()

        units_meta = self.config.property_package.get_metadata().get_derived_units

        self.io_list = io_list = Set(initialize=[0, 1])  # inlet/outlet set

        if hasattr(self.config.property_package,'ion_set'):
            solute_set = self.config.property_package.ion_set
        elif hasattr(self.config.property_package,'solute_set'):
            solute_set = self.config.property_package.solute_set
        else:
            raise ConfigurationError("This NF model was expecting an "
                                     "ion_set or solute_set and did not receive either.")

        solvent_set = self.config.property_package.solvent_set
        solvent_solute_set = solvent_set | solute_set
        phase_list = self.config.property_package.phase_list


        # Build control volume for feed side
        self.feed_side = ControlVolume0DBlock(default={
            "dynamic": False,
            "has_holdup": False,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args})

        self.feed_side.add_state_blocks(
            has_phase_equilibrium=False)

        self.feed_side.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=True)

        self.feed_side.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # Make indexed stateblock and separate stateblock for permeate-side and permeate outlet, respectively.
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False  # these blocks are not inlets

        # Add permeate block
        self.permeate_side = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            io_list,
            doc="Material properties of permeate along permeate channel",
            default=tmp_dict)
        self.mixed_permeate = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of mixed permeate exiting the module",
            default=tmp_dict)

        # Add Ports
        self.add_inlet_port(name='inlet', block=self.feed_side)
        self.add_outlet_port(name='retentate', block=self.feed_side)
        self.add_port(name='permeate', block=self.mixed_permeate)

        # Membrane interface: indexed state block
        self.feed_side.properties_interface = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            io_list,
            doc="Material properties of feed-side membrane interface",
            default=tmp_dict)
        # Pore entrance: indexed state block
        self.pore_entrance = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            io_list,
            doc="Fluid properties within the membrane pore entrance",
            default=tmp_dict)
        # Pore exit: indexed state block
        self.pore_exit = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            io_list,
            doc="Fluid properties within the membrane pore exit",
            default=tmp_dict)

        # References for control volume
        # pressure change
        if (self.config.has_pressure_change is True and
                self.config.momentum_balance_type != 'none'):
            self.deltaP = Reference(self.feed_side.deltaP)

        ###############################################################################################################
        # Variables
        ###############################################################################################################
        #1. Component mole flux, J, DOF=Nj*2 for inlet/outlet
        self.flux_mol_phase_comp = Var(
            self.flowsheet().config.time,
            io_list,
            phase_list,
            solvent_solute_set,
            initialize=(lambda b,t,x,p,j : 2.78e-2 if j in solvent_set else 4e-6), #TODO: divided mass solvent by .018 and solute by .25
            bounds=lambda b,t,x,p,j : (5e-5, 1.5) if j in solvent_set else (4e-12, 1e-2), #TODO: keep checking these
            domain=NonNegativeReals,
            units=units_meta('amount')*units_meta('length')**-2*units_meta('time')**-1,
            doc='Component molar flux at inlet and outlet of membrane')

        # 2. Pore radius, rp, DOF = 1
        self.radius_pore = Var(
            initialize=0.5e-9, #TODO: revisit
            domain=NonNegativeReals,
            units=units_meta('length'),
            doc='Membrane pore radius')

        #3. Effective thickness of membrane, Ak, DOF=1
        self.membrane_thickness_effective = Var(
            initialize=1.33e-6,     # Value used by Labban et al., 2017
            domain=NonNegativeReals,
            units=units_meta('length'),
            doc='Effective membrane thickness')

        #4. Effective thickness of membrane, Ak, DOF=1
        self.membrane_charge_density = Var(
            self.flowsheet().config.time,
            initialize=-50, # near value used in Roy et al.
            domain=Reals,
            units=pyunits.mol*pyunits.m**-3,
            doc='Membrane charge density')
        self.dielectric_constant_pore = Var(
            self.flowsheet().config.time,
            initialize=42, # near value used in Roy et al.
            bounds=(1, None),
            units=pyunits.dimensionless, # TODO: revisit bounds/domain
            doc='Pore dielectric constant')
        self.electric_potential = Var(
            self.flowsheet().config.time,
            io_list,
            ['pore_entrance','pore_exit', 'permeate'],
            initialize=0.1, #TODO:revisit
            domain=Reals,
            bounds=(-1.001, 1.001),
            units=pyunits.V,
            doc='Electric potential of pore entrance/exit, and permeate')
        self.electric_potential_grad_feed_interface = Var(
            self.flowsheet().config.time,
            io_list,
            initialize=1, #TODO: revisit
            domain=Reals,
            units= pyunits.V*pyunits.m**-1, # TODO: revisit- Geraldes and Alves give unitless while Roy et al. give V/m
            doc='Electric potential gradient of feed-membrane interface')
        self.Kf_comp = Var(
            self.flowsheet().config.time,
            self.io_list,
            solute_set,
            initialize=5e-5,
            bounds=(1e-8, 1e-3),
            domain=NonNegativeReals,
            units=units_meta('length') * units_meta('time') ** -1,
            doc='Component mass transfer coefficient in feed channel at inlet and outlet')
        self.rejection_intrinsic_phase_comp = Var(
            self.flowsheet().config.time,
            phase_list,
            solute_set,
            initialize=0.1,
            bounds=(-1.001, 1.001),
            units=pyunits.dimensionless,
            doc='Intrinsic solute rejection')
        self.area = Var(
            initialize=50,
            bounds=(1e-2, 1e3),
            domain=NonNegativeReals,
            units=units_meta('length') ** 2,
            doc='Membrane area')

        self.recovery_vol_phase = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            initialize=0.5,
            bounds=(1e-3, 1 - 1e-6),
            units=pyunits.dimensionless,
            doc='Volumetric-based recovery')
        if self.config.mass_transfer_coefficient == MassTransferCoefficient.spiral_wound:
            self.length = Var(
                initialize=10,
                bounds=(0.1, 5e2),
                domain=NonNegativeReals,
                units=units_meta('length'),
                doc='Effective membrane length')
            self.width = Var(
                initialize=5,
                bounds=(0.1, 5e2),
                domain=NonNegativeReals,
                units=units_meta('length'),
                doc='Effective feed-channel width')
            self.channel_height = Var(
                initialize=1e-3,
                domain=NonNegativeReals,
                bounds=(1e-4, 5e-3),
                units=units_meta('length'),
                doc='Feed channel height')
            self.spacer_porosity = Var(
                initialize=0.95,
                bounds=(0.1, 1.001),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc='Feed-channel spacer porosity')
            self.velocity = Var(
                self.flowsheet().config.time,
                self.io_list,
                initialize=0.5,
                bounds=(1e-3, 2),
                domain=NonNegativeReals,
                units=units_meta('length') / units_meta('time'),
                doc="Crossflow velocity in feed channel at inlet and outlet")
            self.N_Sc_comp = Var(
                self.flowsheet().config.time,
                self.io_list,
                solute_set,
                initialize=5e2,
                bounds=(1e2, 2e3),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Schmidt number at inlet and outlet")
            self.N_Pe_comp = Var(
                self.flowsheet().config.time,
                self.io_list,
                solute_set,
                initialize=1e5,
                bounds=(5e3, None),  # TODO:unsure of value ranges at the moment
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Peclet number at inlet and outlet")
            self.spacer_mixing_efficiency = Var(
                initialize=0.5,
                # bounds=(1e2, 2e3),  # TODO:unsure of value ranges at the moment- since this is efficiency, assuming 0 -1
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Mixing efficiency of spacer net")
            self.spacer_mixing_length = Var(
                initialize=0.6,
                # bounds=(1e2, 5e3),  #TODO:unsure of value ranges at the moment
                domain=NonNegativeReals,
                units=units_meta('length'),
                doc="Characteristic length of spacer")
        ###############################################################################################################
        # Expressions
        ###############################################################################################################
        # Make expressions that don't depend on any variables
        self._make_expressions()

        @self.Expression(self.flowsheet().config.time,
                         io_list,
                         solute_set,
                         doc="Donnan exclusion contribution to partitioning on feed side")
        def partition_factor_donnan_comp_feed(b, t, x, j):
            return (exp(-b.feed_side.properties_in[t].charge_comp[j] * Constants.faraday_constant
                    / (Constants.gas_constant * b.pore_entrance[t, x].temperature)
                    * b.electric_potential[t, x, 'pore_entrance']))

        @self.Expression(self.flowsheet().config.time,
                         io_list,
                         solute_set,
                         doc="Donnan exclusion contribution to partitioning on permeate side")
        def partition_factor_donnan_comp_permeate(b, t, x, j):
            return (exp(-b.feed_side.properties_in[t].charge_comp[j] * Constants.faraday_constant
                    / (Constants.gas_constant * b.pore_exit[t, x].temperature)
                    * (b.electric_potential[t, x, 'pore_exit'] - b.electric_potential[t, x, 'permeate'])))

        # Volumetric Water Flux at inlet and outlet ------------------------------------#
        @self.Expression(self.flowsheet().config.time,
                         io_list,
                         doc="Volumetric water flux at inlet and outlet")
        def flux_vol_water(b, t, x):
            prop = b.feed_side.properties_in[t]
            return b.flux_mol_phase_comp[t, x, 'Liq', 'H2O'] * prop.mw_comp['H2O'] / prop.dens_mass_solvent

        # Average Volumetric Water Flux ------------------------------------#
        @self.Expression(self.flowsheet().config.time,
                         doc="Average volumetric water flux")
        def flux_vol_water_avg(b, t):
            return sum(b.flux_vol_water[t, x] for x in io_list) / len(io_list)

        # Average mole flux of each component ------------------------------------#
        @self.Expression(self.flowsheet().config.time,
                         phase_list,
                         solvent_solute_set,
                         doc="Average molar component flux")
        def flux_mol_phase_comp_avg(b, t, p, j):
            return sum(b.flux_mol_phase_comp[t, x, p, j] for x in io_list) / len(io_list)

        # Average concentration inside the membrane------------------------------------#
        @self.Expression(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solvent_solute_set,
                         doc="Average molar concentration inside the membrane")
        def conc_mol_phase_comp_pore_avg(b, t, x, p, j):
            return (b.pore_entrance[t, x].conc_mol_phase_comp[p, j]
                    + b.pore_exit[t, x].conc_mol_phase_comp[p, j]) / len(io_list)

        # TODO - no relationship described between mixing length and spacer mixing efficiency with spacer porosity.
        #  Need effective cross-sectional area for velocity at inlet AND outlet. Assuming spacer porosity as an
        #  additional variable in the model that is independent of aforementioned parameters which are used in
        #  the mass transfer coefficient calculation for spiral wound modules. Revisit later.

        if self.config.mass_transfer_coefficient == MassTransferCoefficient.spiral_wound:
            # Cross sectional area ------------------------------------#
            @self.Expression(doc="Cross-sectional area")
            def area_cross(b):
                return b.channel_height * b.width * b.spacer_porosity

        ################################################################################################################
        # Constraints
        ################################################################################################################
        if self.config.mass_transfer_coefficient == MassTransferCoefficient.spiral_wound:
            # 0. Membrane area
            @self.Constraint(doc="Membrane area")
            def eq_area(b):
                return b.area == b.length * b.width

        # 1. Feed-solution/membrane equilibrium, DOF= Nj * 2 for inlet/outlet
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc="Interfacial partitioning at feed side of membrane")
        def eq_interfacial_partitioning_feed(b, t, x, p, j):
            return (b.pore_entrance[t, x].act_coeff_phase_comp[p, j] * b.pore_entrance[t, x].conc_mol_phase_comp[p, j]
                    / (b.feed_side.properties_interface[t, x].act_coeff_phase_comp[p, j]
                    * b.feed_side.properties_interface[t, x].conc_mol_phase_comp[p, j])
                    ==
                    b.partition_factor_steric_comp[t, j]
                    * b.partition_factor_born_solvation_comp[t, j]
                    * b.partition_factor_donnan_comp_feed[t, x, j]
                    )

        # 2. Permeate solution/membrane equilibrium, DOF= Nj * 2 for inlet/outlet
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc="Interfacial partitioning at permeate side of membrane")
        def eq_interfacial_partitioning_permeate(b, t, x, p, j):
            return (b.pore_exit[t, x].act_coeff_phase_comp[p, j] * b.pore_exit[t, x].conc_mol_phase_comp[p, j]
                    / (b.permeate_side[t, x].act_coeff_phase_comp[p, j]
                    * b.permeate_side[t, x].conc_mol_phase_comp[p, j])
                    ==
                    b.partition_factor_steric_comp[t, j]
                    * b.partition_factor_born_solvation_comp[t, j]
                    * b.partition_factor_donnan_comp_permeate[t, x, j])

        # 3. Feed-solution/membrane electroneutrality, DOF=1 *2 for inlet/outlet: DOF= 2
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         doc="Electroneutrality at feed-side membrane interface")
        def eq_electroneutrality_interface(b, t, x, p):
            return (sum(b.feed_side.properties_interface[t, x].conc_mol_phase_comp[p, j] *
                    b.feed_side.properties_interface[t, x].charge_comp[j] for j in solute_set)
                    == 0)
            #todo: tolerance should just be 0

        # 4. Charge balance inside the membrane, DOF=N nodes across membrane thickness *2 for inlet/outlet: N=2, DOF=4
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         ['pore_entrance', 'pore_exit'],
                         phase_list,
                         doc="Electroneutrality within membrane pore")
        def eq_electroneutrality_pore(b, t, x, y, p):
            if y == 'pore_entrance':
                pore_loc = b.pore_entrance[t, x]
            elif y == 'pore_exit':
                pore_loc = b.pore_exit[t, x]
            return (sum(pore_loc.conc_mol_phase_comp[p, j]
                    * pore_loc.charge_comp[j] for j in solute_set)
                    + b.membrane_charge_density[t] == 0)

        # 4. Permeate electroneutrality, DOF=1 *2 for inlet/outlet:  DOF=2
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         doc="Electroneutrality in permeate")
        def eq_electroneutrality_permeate(b, t, x, p):
            return (sum(b.permeate_side[t, x].conc_mol_phase_comp[p, j] *
                        b.permeate_side[t, x].charge_comp[j] for j in solute_set) == 0)

        # 5. Water flux via Hagen-Poiseuille relationship, DOF= 1 * 2 for inlet/outlet: DOF= 2
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         doc="Hagen-Poiseuille relationship for water flux across membrane")
        def eq_water_flux(b, t, x, p):
            if x == 0:
                prop_feed = b.feed_side.properties_in[t]
            elif x == 1:
                prop_feed = b.feed_side.properties_out[t]
            prop_perm = b.permeate_side[t, x]
            prop_feed_inter = b.feed_side.properties_interface[t, x]
            return (b.flux_vol_water[t, x] ==
                    (prop_feed.pressure - prop_perm.pressure
                     -(prop_feed_inter.pressure_osm - prop_perm.pressure_osm)
                     )
                    * (b.radius_pore ** 2)
                    / (8 * prop_feed.visc_d_phase[p] * b.membrane_thickness_effective)
                    )

        # 6. Unhindered mass transfer; Js,i=Jw*cp,i; DOF= Nj * 2 for inlet/outlet
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc="Solute flux as function of solvent flux")
        def eq_solute_solvent_flux(b, t, x, p, j):
            return (b.flux_mol_phase_comp[t, x, p, j] ==
                    b.flux_vol_water[t, x] * b.permeate_side[t, x].conc_mol_phase_comp[p, j])

        # TESTING PROBLEMATIC CONSTRAINT RESULTING IN ERRONEOUSLY LOW REJECTION:
        @self.Expression(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc="Diffusive transport across membrane pore")
        def diffusive_term(b, t, x, p, j):
            return (- b.diffus_pore_comp[t, j]
                    * (b.pore_exit[t, x].conc_mol_phase_comp[p, j] - b.pore_entrance[t, x].conc_mol_phase_comp[p, j])
                    / b.membrane_thickness_effective)

        @self.Expression(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc="Convective transport across membrane pore")
        def convective_term(b, t, x, p, j):
            return (b.hindrance_factor_convective_comp[t, j]
                    * b.conc_mol_phase_comp_pore_avg[t, x, p, j]
                    * b.flux_vol_water[t, x])

        @self.Expression(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc="Electromigrative transport across membrane pore")
        def electromigration_term(b, t, x, p, j):
            return (- b.feed_side.properties_in[t].charge_comp[j]
                    * b.conc_mol_phase_comp_pore_avg[t, x, p, j]
                    * b.diffus_pore_comp[t, j]
                    * Constants.faraday_constant / (Constants.gas_constant * b.feed_side.properties_in[t].temperature)
                    * (b.electric_potential[t, x, 'pore_exit'] - b.electric_potential[t, x, 'pore_entrance'])
                    / b.membrane_thickness_effective)

        # 7. Extended Nernst Planck equation, DOF= Nj * 2 for inlet/outlet
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc="Solute flux within pore domain")
        def eq_solute_flux_pore_domain(b, t, x, p, j):
            return (b.flux_mol_phase_comp[t, x, p, j] ==
                    b.diffusive_term[t, x, p, j]
                    + b.convective_term[t, x, p, j]
                    + b.electromigration_term[t, x, p, j]
                    )

        # 8. Feed-solution/membrane mass transfer resistance, DOF= Nj * 2 for inlet/outlet
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         solute_set,
                         doc='Feed-interface mass transfer resistance accounting for concentration polarization')
        def eq_solute_flux_concentration_polarization(b, t, x, p, j):
            if x == 0:
                bulk = b.feed_side.properties_in[t]
            elif x:
                bulk = b.feed_side.properties_out[t]
            interface = b.feed_side.properties_interface[t, x]
            return (b.flux_mol_phase_comp[t, x, p, j] ==
                    - b.Kf_comp[t, x, j]
                    * (interface.conc_mol_phase_comp[p, j]
                       - bulk.conc_mol_phase_comp[p, j])
                    + b.flux_vol_water[t, x]
                    * interface.conc_mol_phase_comp[p, j]
                    - interface.charge_comp[j]
                    * interface.conc_mol_phase_comp[p, j]
                    * interface.diffus_phase_comp[p, j]
                    * Constants.faraday_constant
                    / Constants.gas_constant
                    / interface.temperature
                    * b.electric_potential_grad_feed_interface[t, x])

        # 9. Isothermal conditions at permeate inlet/outlet, DOF= 1*2 for inlet/outlet
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         doc="Isothermal assumption for permeate")
        def eq_permeate_isothermal(b, t, x):
            return b.feed_side.properties_in[t].temperature == \
                   b.permeate_side[t, x].temperature

        # 10. Isothermal conditions at feed/membrane interface, DOF= 1*2 for inlet/outlet
        @self.feed_side.Constraint(self.flowsheet().config.time,
                                   io_list,
                                   doc="Isothermal assumption for feed-membrane interface")
        def eq_feed_interface_isothermal(b, t, x):
            return b.properties_in[t].temperature == \
                   b.properties_interface[t, x].temperature

        # 11. Isobaric conditions at permeate side, DOF= 1*2 for inlet/outlet
        # TOdo: mixed permeate pressure is currently unused variable, but fixing its value satisfies this constraint
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         doc="Isobaric permeate pressure")
        def eq_pressure_permeate_io(b, t, x):
            return b.permeate_side[t, x].pressure == b.mixed_permeate[t].pressure

        # 12. Javg * area = -dMf
        @self.Constraint(self.flowsheet().config.time,
                         phase_list,
                         solvent_solute_set,
                         doc="Component mass transfer from feed")
        def eq_mass_transfer_feed(b, t, p, j):
            if b.feed_side.properties_in[0].get_material_flow_basis() == MaterialFlowBasis.molar:
                return b.flux_mol_phase_comp_avg[t, p, j] * b.area == -b.feed_side.mass_transfer_term[t, p, j]

        # 13. Mass transfer equal to permeate flow terms; mole_flow,perm final = -dMf = Javg * area
        @self.Constraint(self.flowsheet().config.time,
                         phase_list,
                         solvent_solute_set,
                         doc="Permeate production/average mass transfer constraint")
        def eq_permeate_production(b, t, p, j):
            if b.mixed_permeate[0].get_material_flow_basis() == MaterialFlowBasis.molar:
                return (b.mixed_permeate[t].get_material_flow_terms(p, j)
                        == b.flux_mol_phase_comp_avg[t, p, j] * b.area)

        # 14. Volumetric recovery rate
        @self.Constraint(self.flowsheet().config.time,
                         phase_list)
        def eq_recovery_vol_phase(b, t, p):
            return (b.recovery_vol_phase[t, p]
                    * b.feed_side.properties_in[t].flow_vol_phase[p]
                    == b.mixed_permeate[t].flow_vol_phase[p])

        # 15. Intrinsic rejection rate
        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         solute_set,
                         doc="Intrinsic solute rejection")
        def eq_rejection_intrinsic_phase_comp(b, t, p, j):
            return (b.rejection_intrinsic_phase_comp[t, p, j] ==
                    1 - b.mixed_permeate[t].conc_mol_phase_comp[p, j]
                    / b.feed_side.properties_interface[t, 0].conc_mol_phase_comp[p, j])

        if self.config.mass_transfer_coefficient == MassTransferCoefficient.spiral_wound:
            # 16. Mass transfer coefficient
            @self.Constraint(self.flowsheet().config.time,
                             io_list,
                             solute_set,
                             doc="Mass transfer coefficient")
            def eq_Kf_comp(b, t, x, j):
                bulk_diff = b.feed_side.properties_in[t].diffus_phase_comp['Liq', j]
                return (b.Kf_comp[t, x, j] ==
                        0.753
                        * (b.spacer_mixing_efficiency/(2 - b.spacer_mixing_efficiency)) ** 0.5
                        * (2 * bulk_diff / b.channel_height) * b.N_Sc_comp[t, x, j] ** (-1/6)
                        * (b.N_Pe_comp[t, x, j] * b.channel_height / (2 * b.spacer_mixing_length)) ** 0.5)
                #TODO: NOTE--- error in MIT paper; 1/2 of channel height should be in numerator

            # 17. Schmidt number calculation
            @self.Constraint(self.flowsheet().config.time,
                             io_list,
                             solute_set,
                             doc="Schmidt number equation")
            def eq_N_Sc_comp(b, t, x, j):
                if not x:
                    prop_io = b.feed_side.properties_in[t]
                elif x:
                    prop_io = b.feed_side.properties_out[t]
                return (b.N_Sc_comp[t, x, j] * prop_io.dens_mass_phase['Liq'] * prop_io.diffus_phase_comp['Liq', j] ==
                        prop_io.visc_d_phase['Liq'])

            # 18. Peclet number calculation
            @self.Constraint(self.flowsheet().config.time,
                             io_list,
                             solute_set,
                             doc="Peclet number equation")
            def eq_N_Pe_comp(b, t, x, j):
                bulk_diff = b.feed_side.properties_in[t].diffus_phase_comp['Liq', j]
                return b.N_Pe_comp[t, x, j] == 2 * b.channel_height * b.velocity[t, x] / bulk_diff

            # 19. Crossflow velocity at inlet and outlet
            @self.Constraint(self.flowsheet().config.time,
                             self.io_list,
                             doc="Crossflow velocity constraint")
            def eq_velocity(b, t, x):
                if not x:
                    prop_io = b.feed_side.properties_in[t]
                elif x:
                    prop_io = b.feed_side.properties_out[t]
                return b.velocity[t, x] * b.area_cross == prop_io.flow_vol_phase['Liq']

        # TODO: seems stale since temperature unused at pore entrance/exit- confirm+remove;
        #  1/17/22: after including temp variables for pore in interfacial equilib eqns, this is relevant
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         ['pore_entrance', 'pore_exit'],
                         doc="Isothermal assumption for pore inlet/outlet")
        def eq_pore_isothermal(b, t, x, y):
            if y == 'pore_entrance':
                prop = b.pore_entrance[t, x]
            elif y == 'pore_exit':
                prop = b.pore_exit[t, x]
            return b.feed_side.properties_in[t].temperature == \
                   prop.temperature

        # Experimental Constraint with new density calculation in prop package-- temp equality in permeate
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         doc="Isothermal assumption for mixed permeate")
        def eq_permeate_isothermal_mixed(b, t, x):
            return b.feed_side.properties_in[t].temperature == \
                   b.mixed_permeate[t].temperature

        # Experimental constraint: noticed feed outlet temp didn't match inlet
        @self.feed_side.Constraint(self.flowsheet().config.time,
                                   doc="Isothermal assumption for feed-outlet")
        def eq_feed_isothermal(b, t):
            return b.properties_in[t].temperature == \
                   b.properties_out[t].temperature

        # Experimental constraint
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         doc="Equal volumetric flow for pore exit and permeate at inlet and outlet")
        def eq_equal_flow_vol_pore_exit_permeate(b, t, x):
            return b.pore_entrance[t, x].flow_vol_phase['Liq'] ==\
                   b.pore_exit[t, x].flow_vol_phase['Liq']

        # Experimental constraint
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         doc="Volumetric flow at permeate of inlet and outlet equal to mixed permeate")
        def eq_equal_flow_vol_permeate(b, t, x):
            return b.permeate_side[t, x].flow_vol_phase['Liq'] ==\
                   b.mixed_permeate[t].flow_vol_phase['Liq']

        # Experimental constraint: Electroneutrality of final permeate
        @self.Constraint(self.flowsheet().config.time,
                         phase_list,
                         doc="Electroneutrality in mixed permeate")
        def eq_electroneutrality_mixed_permeate(b, t, p):
            return (sum(b.mixed_permeate[t].conc_mol_phase_comp[p, j] *
                        b.mixed_permeate[t].charge_comp[j] for j in solute_set) == 0)
        #
        # Experimental constraint: feed electroneutrality
        @self.Constraint(self.flowsheet().config.time,
                         phase_list,
                         doc="Electroneutrality at feed outlet")
        def eq_electroneutrality_feed_outlet(b, t, p):
            prop = b.feed_side.properties_out[t]
            return (sum(prop.conc_mol_phase_comp[p, j] *
                    prop.charge_comp[j] for j in solute_set)
                    == 0)

        # Experimental constraint: electroneutrality inside membrane
        @self.Constraint(self.flowsheet().config.time,
                         io_list,
                         phase_list,
                         doc="Electroneutrality inside the membrane")
        def eq_electroneutrality_in_membrane(b, t, x, p):
            return (sum(b.conc_mol_phase_comp_pore_avg[t, x, p, j] *
                    b.feed_side.properties_in[t].charge_comp[j] for j in solute_set)
                    + b.membrane_charge_density[t]
                    == 0)

    def _make_expressions(self):
        solute_set = self.config.property_package.solute_set

        # Stokes radius to membrane pore radius ratio (for each solute)
        @self.Expression(self.flowsheet().config.time,
                         solute_set,
                         doc="Ratio of stokes radius to membrane pore radius equation")
        def lambda_comp(b, t, j):
            return smooth_min(1, b.feed_side.properties_in[t].radius_stokes_comp[j] / b.radius_pore)

        @self.Expression(self.flowsheet().config.time,
                         solute_set,
                         doc="Diffusive hindered transport coefficient")
        def hindrance_factor_diffusive_comp(b, t, j):
            eps = 1e-8
            return Expr_if(b.lambda_comp[t, j] > 0.95,
                           0.984 * ((1 - b.lambda_comp[t, j]) / b.lambda_comp[t, j]) ** (5 / 2),
                           (1 + 9. / 8. * b.lambda_comp[t, j] * log(b.lambda_comp[t, j])
                            - 1.56034 * b.lambda_comp[t, j]
                            + 0.528155 * b.lambda_comp[t, j] ** 2
                            + 1.91521 * b.lambda_comp[t, j] ** 3
                            - 2.81903 * b.lambda_comp[t, j] ** 4
                            + 0.270788 * b.lambda_comp[t, j] ** 5
                            + 1.10115 * b.lambda_comp[t, j] ** 6
                            - 0.435933 * b.lambda_comp[t, j] ** 7) /
                           (1 - b.lambda_comp[t, j] + eps) ** 2,
                           )
                # Relationship used by Geraldes & Alves
                # (1 - 2.3 * b.lambda_comp[t, j]
                #     + 1.154 * b.lambda_comp[t, j] ** 2
                #     + 0.224 * b.lambda_comp[t, j] ** 3
                #     )

        @self.Expression(self.flowsheet().config.time,
                         solute_set,
                         doc="Pore diffusion coefficient")
        def diffus_pore_comp(b, t, j):
            return b.hindrance_factor_diffusive_comp[t, j] * b.feed_side.properties_in[t].diffus_phase_comp['Liq', j]

        @self.Expression(self.flowsheet().config.time,
                         solute_set,
                         doc="Convective hindered transport coefficient")
        def hindrance_factor_convective_comp(b, t, j):
            return ((1 + 3.867 * b.lambda_comp[t, j]
                     - 1.907 * b.lambda_comp[t, j] ** 2
                     - 0.834 * b.lambda_comp[t, j] ** 3)
                    / (1 + 1.867 * b.lambda_comp[t, j] - 0.741 * b.lambda_comp[t, j]**2))

        @self.Expression(self.flowsheet().config.time,
                         solute_set,
                         doc="Steric partitioning factor")
        def partition_factor_steric_comp(b, t, j):
            return (1 - b.lambda_comp[t, j]) ** 2

        @self.Expression(self.flowsheet().config.time,
                         solute_set,
                         doc="Gibbs free energy of solvation for each ion")
        def gibbs_solvation_comp(b, t, j):
            return (b.feed_side.properties_in[t].charge_comp[j] ** 2
                    * Constants.elemental_charge ** 2
                    / (8 * Constants.pi
                       * Constants.vacuum_electric_permittivity
                       * b.feed_side.properties_in[t].radius_stokes_comp[j])
                    * (1 / b.feed_side.properties_in[t].dielectric_constant - 1 / b.dielectric_constant_pore[t]))

        @self.Expression(self.flowsheet().config.time,
                         solute_set,
                         doc="Born solvation contribution to partitioning")
        def partition_factor_born_solvation_comp(b, t, j):
            return (-b.gibbs_solvation_comp[t, j]
                    / (Constants.boltzmann_constant * b.feed_side.properties_in[t].temperature))

    def initialize_build(
            self,
            initialize_guess=None,
            state_args=None,
            outlvl=idaeslog.NOTSET,
            solver=None,
            optarg=None,
            fail_on_warning=False,
            ignore_dof=False,
            automate_rescale=True):

        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            initialize_guess : a dict of guesses for ....
                   #TODO: revise as appropriate
                   solvent_recovery, solute_recovery,
                   and cp_modulus. These guesses offset the initial values
                   for the retentate, permeate, and membrane interface
                   state blocks from the inlet feed
                   (default =
                   {'deltaP': -1e4,
                   'solvent_recovery': 0.5,
                   'solute_recovery': 0.01,
                   'cp_modulus': 1.1})
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)
            fail_on_warning : boolean argument to fail or only produce  warning upon unsuccessful solve (default=False)
            ignore_dof : boolean argument to ignore when DOF != 0 (default=False)
            automate_rescale: boolean argument to automatically rescale poorly scaled vars
        Returns:
            None
        """

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Extract initial state of inlet feed
        source = self.feed_side.properties_in[self.flowsheet().config.time.first()]
        state_args = self._get_state_args(source, self.mixed_permeate[0], initialize_guess, state_args)

        # ---------------------------------------------------------------------
        # Initialize holdup block
        flags_feed_side = self.feed_side.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['feed_side'],
            hold_state=True
        )
        init_log.info_high("Initialization Step 1 Complete.")
        if not ignore_dof:
            check_dof(self, fail_flag=fail_on_warning, logger=init_log)
        # ---------------------------------------------------------------------
        # Initialize other state blocks based on properties at
        # inlet state block
        self.feed_side.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['retentate'],)
        self.feed_side.properties_interface.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['interface_in'], )
        self.permeate_side.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['permeate'], )
        self.mixed_permeate.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['permeate'], )
        self.pore_entrance.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['interface_in'],)
        self.pore_exit.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['permeate'],)
        init_log.info_high("Initialization Step 2 Complete.")

        # Double-check for poorly scaled variables after state block initialization
        # and rescale them so that scaled variable values = 1:
        if automate_rescale:
            badly_scaled_vars = list(iscale.badly_scaled_var_generator(self))
            if len(badly_scaled_vars) > 0:
                init_log.warn(f"{len(badly_scaled_vars)} poorly scaled "
                              f"variable(s) will be rescaled so that each scaled variable value = 1")
                self._automate_rescale_variables()
        # ---------------------------------------------------------------------
        # Solve unit with deactivated constraint
        # self.eq_solute_solvent_flux.deactivate()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
            if not check_optimal_termination(res):
                init_log.warn("Trouble solving NanofiltrationDSPMDE0D unit model with deactivated constraint.")
        # ---------------------------------------------------------------------
        # Solve unit
        self.eq_solute_solvent_flux.activate()
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
            if not check_optimal_termination(res):
                init_log.warn("Trouble solving NanofiltrationDSPMDE0D unit model. Trying one more time.")
                res = opt.solve(self, tee=slc.tee)
                if not check_optimal_termination(res):
                    raise InitializationError('The property package failed to solve during initialization.')
        check_solve(res, checkpoint='Solve in Initialization Step 3', logger=init_log, fail_flag=fail_on_warning)
        # Release Inlet state
        self.feed_side.release_state(flags_feed_side, outlvl)
        # Rescale any badly scaled vars
        if automate_rescale:
            badly_scaled_vars = list(iscale.badly_scaled_var_generator(self))
            if len(badly_scaled_vars) > 0:
                init_log.warn(f"After solve: {len(badly_scaled_vars)} poorly scaled "
                              f"variable(s) will be rescaled so that each scaled variable value = 1")
            self._automate_rescale_variables()
        init_log.info(f"Initialization Complete: {idaeslog.condition(res)}")

    def _get_performance_contents(self, time_point=0):
        # TODO: replace 0 with time_point
        var_dict = {}
        expr_dict = {}
        var_dict["Volumetric Recovery Rate"] = self.recovery_vol_phase[time_point, 'Liq']
        var_dict["Membrane Area"] = self.area
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]
        # Volume flowrates
        if self.feed_side.properties_in[time_point].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ feed inlet'] = self.feed_side.properties_in[time_point].flow_vol_phase['Liq']
        if self.feed_side.properties_interface[time_point, 0].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ inlet interface'] = self.feed_side.properties_interface[time_point, 0].flow_vol_phase['Liq']
        if self.pore_entrance[time_point, 0].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ inlet pore entrance'] = self.pore_entrance[time_point, 0].flow_vol_phase['Liq']
        if self.pore_exit[time_point, 0].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ inlet pore exit'] = self.pore_exit[time_point, 0].flow_vol_phase['Liq']
        if self.permeate_side[time_point, 0].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ inlet permeate'] = self.permeate_side[time_point, 0].flow_vol_phase['Liq']

        if self.feed_side.properties_out[time_point].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ feed outlet'] = self.feed_side.properties_out[time_point].flow_vol_phase['Liq']
            if self.feed_side.properties_interface[time_point, 1].is_property_constructed('flow_vol_phase'):
                var_dict[f'Volumetric flow rate @ outlet interface'] = self.feed_side.properties_interface[time_point, 1].flow_vol_phase['Liq']
        if self.pore_entrance[time_point, 1].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ outlet pore entrance'] = self.pore_entrance[time_point, 1].flow_vol_phase['Liq']
        if self.pore_exit[time_point, 1].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ outlet pore exit'] = self.pore_exit[time_point, 1].flow_vol_phase['Liq']
        if self.permeate_side[time_point, 0].is_property_constructed('flow_vol_phase'):
            var_dict[f'Volumetric flow rate @ outlet permeate'] = self.permeate_side[time_point, 1].flow_vol_phase['Liq']
        var_dict[f'Volumetric flow rate @ final permeate'] = self.mixed_permeate[time_point].flow_vol_phase['Liq']

        expr_dict['Average Volumetric Flux (LMH)'] = self.flux_vol_water_avg[time_point] *3.6e6
        for j in self.config.property_package.component_list:
            expr_dict[f'Average Mole FLux of {j} '] = self.flux_mol_phase_comp_avg[time_point, 'Liq', j]
            # Molar flowrates
            var_dict[f'Molar flow rate of {j} @ feed inlet'] = \
                self.feed_side.properties_in[time_point].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ feed outlet'] = \
                self.feed_side.properties_out[time_point].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ membrane-interface inlet'] = \
                self.feed_side.properties_interface[time_point, 0].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ membrane-interface outlet'] = \
                self.feed_side.properties_interface[time_point, 1].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ pore entrance, inlet'] = \
                self.pore_entrance[time_point, 0].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ pore entrance, outlet'] = \
                self.pore_entrance[time_point, 1].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ pore exit, inlet'] = \
                self.pore_exit[time_point, 0].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ pore exit, outlet'] = \
                self.pore_exit[time_point, 1].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ permeate, inlet'] = \
                self.permeate_side[time_point, 0].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ permeate, outlet'] = \
                self.permeate_side[time_point, 1].flow_mol_phase_comp['Liq', j]
            var_dict[f'Molar flow rate of {j} @ mixed permeate'] = \
                self.mixed_permeate[time_point].flow_mol_phase_comp['Liq', j]

        for j in self.config.property_package.solute_set:
            expr_dict[f'Stokes radius of {j}'] = self.feed_side.properties_in[time_point].radius_stokes_comp[j]
            expr_dict[f'Stokes:Pore Radius Ratio of {j}'] = self.lambda_comp[time_point, j]
            expr_dict[f'Diffusive Hindrance Factor of {j}'] = self.hindrance_factor_diffusive_comp[time_point, j]
            expr_dict[f'Convective Hindrance Factor of {j}'] = self.hindrance_factor_convective_comp[time_point, j]
            expr_dict[f'Pore Diffusivity of {j}'] = self.diffus_pore_comp[time_point, j]

            expr_dict[f'Gibbs Free Energy of Solvation of {j}'] = self.gibbs_solvation_comp[time_point, j]
            expr_dict[f'Born Solvation Energy Partitioning Factor of {j}'] = self.partition_factor_born_solvation_comp[time_point, j]
            expr_dict[f'Steric Hindrance Partitioning Factor of {j}'] = self.partition_factor_steric_comp[time_point, j]
            expr_dict[f'Donnan Partitioning Factor of {j} @ Feed-side Inlet'] = self.partition_factor_donnan_comp_feed[time_point, 0, j]
            expr_dict[f'Donnan Partitioning Factor of {j} @ Permeate-side Inlet'] = self.partition_factor_donnan_comp_permeate[time_point, 0, j]
            expr_dict[f'Donnan Partitioning Factor of {j} @ Feed-side Outlet'] = self.partition_factor_donnan_comp_feed[time_point, 1, j]
            expr_dict[f'Donnan Partitioning Factor of {j} @ Permeate-side Outlet'] = self.partition_factor_donnan_comp_permeate[time_point, 1, j]

            var_dict[f'Intrinsic Rejection of {j}'] = self.rejection_intrinsic_phase_comp[time_point, 'Liq', j]

            var_dict[f'Molar Concentration of {j} @ Feed Inlet'] = self.feed_side.properties_in[time_point].conc_mol_phase_comp['Liq', j]
            var_dict[f'Molar Concentration of {j} @ Feed Outlet'] = self.feed_side.properties_out[time_point].conc_mol_phase_comp['Liq', j]
            var_dict[f'Molar Concentration of {j} @ Final Permeate'] = \
                self.mixed_permeate[time_point].conc_mol_phase_comp['Liq', j]

            if self.feed_side.properties_in[time_point].is_property_constructed('pressure_osm'):
                var_dict[f"Osmotic Pressure @ Bulk Feed, Inlet (Pa)"] = \
                    self.feed_side.properties_in[time_point].pressure_osm
            if self.feed_side.properties_out[time_point].is_property_constructed('pressure_osm'):
                var_dict[f"Osmotic Pressure @ Bulk Feed, Outlet (Pa)"] = \
                    self.feed_side.properties_out[time_point].pressure_osm

            for x in self.io_list:
                if not x:
                    io = 'Inlet'
                    prop_feed = self.feed_side.properties_in[0]
                elif x:
                    io = 'Outlet'
                    prop_feed = self.feed_side.properties_out[0]

                var_dict[f'Molar Concentration of {j} @ Membrane Interface, {io}'] = \
                    self.feed_side.properties_interface[time_point, x].conc_mol_phase_comp['Liq', j]
                var_dict[f'Molar Concentration of {j} @ Pore Entrance, {io}'] = \
                    self.pore_entrance[time_point, x].conc_mol_phase_comp['Liq', j]
                var_dict[f'Molar Concentration of {j} @ Pore Exit, {io}'] = \
                    self.pore_exit[time_point, x].conc_mol_phase_comp['Liq', j]
                var_dict[f'Molar Concentration of {j} @ Permeate, {io}'] = \
                    self.permeate_side[time_point, x].conc_mol_phase_comp['Liq', j]


                var_dict[f"Osmotic Pressure @ Membrane Interface, {io} (Pa)"] = \
                    self.feed_side.properties_interface[time_point, x].pressure_osm

                var_dict[f"Osmotic Pressure @ Permeate, {io} (Pa)"] = \
                    self.permeate_side[time_point, x].pressure_osm
                expr_dict[f"Net Driving Pressure, {io} (Pa)"] = (prop_feed.pressure
                                                                 - self.permeate_side[0, x].pressure
                                                                 - (self.feed_side.properties_interface[0, x].pressure_osm
                                                                    - self.permeate_side[0, x].pressure_osm))
                var_dict[f'Electric Potential @ Pore Entrance, {io}'] = self.electric_potential[0, x, 'pore_entrance']
                var_dict[f'Electric Potential @ Pore Exit, {io}'] = self.electric_potential[0, x, 'pore_exit']
                var_dict[f'Electric Potential @ Permeate, {io}'] = self.electric_potential[0, x, 'permeate']
                var_dict[f'Electric Potential Gradient @ Feed-Membrane Interface, {io}'] = \
                    self.electric_potential_grad_feed_interface[0, x]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Feed Inlet": self.inlet,
                "Feed Outlet": self.retentate,
                "Permeate Outlet": self.permeate,
            },
            time_point=time_point,
        )

    def get_costing(self, module=None, **kwargs):
        self.costing = Block()
        module.Nanofiltration_costing(self.costing, **kwargs)

    def _get_state_args(self, source, mixed_permeate_properties, initialize_guess, state_args):
        '''
        This method returns state_args (initial guesses for state variable values) to be passed to
        each stateblock during initialization.
        Arguments:
            source : property model containing inlet feed
            mixed_permeate_properties : mixed permeate property block
            initialize_guess : a dict of guesses for solvent_recovery, solute_recovery,
                               and cp_modulus. These guesses offset the initial values
                               for the retentate, permeate, and membrane interface
                               state blocks from the inlet feed
                               (default =
                               {'deltaP': -1e4,
                               'solvent_recovery': 0.5,
                               'solute_recovery': 0.01,
                               'cp_modulus': 1.1})
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for the inlet
                         feed side state block (see documentation of the specific
                         property package).
        '''

        # assumptions
        if initialize_guess is None:
            initialize_guess = {}
        #TODO: enable deltaP guess when pressure drop is added
        if 'deltaP' not in initialize_guess:
            initialize_guess['deltaP'] = 0
        if 'solvent_recovery' not in initialize_guess:
            initialize_guess['solvent_recovery'] = 0.1
        if 'solute_recovery' not in initialize_guess:
            initialize_guess['solute_recovery'] = 0.1
        if 'cp_modulus' not in initialize_guess:
            initialize_guess['cp_modulus'] = 1.1

        if state_args is None:
            state_args = {}
            state_dict = source.define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        if 'flow_mol_phase_comp' not in state_args.keys():
            raise ConfigurationError(f'{self.__class__.__name__} initialization routine expects '
                                     'flow_mol_phase_comp as a state variable. Check '
                                     'that the property package supports this state '
                                     'variable or that the state_args provided to the '
                                     'initialize call includes this state variable')

        # slightly modify initial values for other state blocks
        state_args_retentate = deepcopy(state_args)
        state_args_permeate = deepcopy(state_args)

        state_args_retentate['pressure'] += initialize_guess['deltaP']
        state_args_permeate['pressure'] = mixed_permeate_properties.pressure.value
        for j in self.config.property_package.solvent_set:
            state_args_retentate['flow_mol_phase_comp'][('Liq', j)] *= (1 - initialize_guess['solvent_recovery'])
            state_args_permeate['flow_mol_phase_comp'][('Liq', j)] *= initialize_guess['solvent_recovery']
        for j in self.config.property_package.solute_set:
            state_args_retentate['flow_mol_phase_comp'][('Liq', j)] *= (1 - initialize_guess['solute_recovery'])
            state_args_permeate['flow_mol_phase_comp'][('Liq', j)] *= initialize_guess['solute_recovery']

        state_args_interface_in = deepcopy(state_args)
        state_args_interface_out = deepcopy(state_args_retentate)

        for j in self.config.property_package.solute_set:
            state_args_interface_in['flow_mol_phase_comp'][('Liq', j)] *= initialize_guess['cp_modulus']
            state_args_interface_out['flow_mol_phase_comp'][('Liq', j)] *= initialize_guess['cp_modulus']

        return {'feed_side': state_args,
                'retentate': state_args_retentate,
                'permeate': state_args_permeate,
                'interface_in': state_args_interface_in,
                'interface_out': state_args_interface_out,
               }

    # stateblock properties need to rescale solute values by a larger factor
    def _rescale_sb_variable(self, var, factor=100):
        if var not in self._sb_scaled_properties:
            sf = iscale.get_scaling_factor(var)
            iscale.set_scaling_factor(var, sf * factor)
            self._sb_scaled_properties.add(var)

    # automatically rescale poorly scaled variables by setting a new scaling factor
    # which multiples a variable value by the old scaling factor divided by the poorly scaled (resulting) value,
    # bringing the new scaled value to 1. Providing a rescale_factor would just multiply that factor by 1.
    def _automate_rescale_variables(self, rescale_factor=None):
        if rescale_factor is None:
            rescale_factor = 1
        for var, sv in iscale.badly_scaled_var_generator(self):
            if iscale.get_scaling_factor(var) is None:
                print(f'{var} is missing a scaling factor')
                continue
            sf = iscale.get_scaling_factor(var)
            iscale.set_scaling_factor(var, sf / sv * rescale_factor)
            iscale.calculate_scaling_factors(self)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for k in ('ion_set', 'solute_set'):
            if hasattr(self.config.property_package, k):
                solute_set = getattr(self.config.property_package, k)
                break

        # setting scaling factors for variables
        if iscale.get_scaling_factor(self.radius_pore) is None:
            iscale.set_scaling_factor(self.radius_pore, 1e10)
        if iscale.get_scaling_factor(self.membrane_thickness_effective) is None:
            iscale.set_scaling_factor(self.membrane_thickness_effective, 1e7)

        # setting scaling factors for variables
        # these variables should have user input, if not there will be a warning
        if iscale.get_scaling_factor(self.area) is None:
            sf = iscale.get_scaling_factor(self.area, default=1, warning=True)
            iscale.set_scaling_factor(self.area, sf)

        for (t, x, y), v in self.electric_potential.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, 1e4)

        for (t, x), v in self.electric_potential_grad_feed_interface.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, 1e-2)

        # these variables do not typically require user input,
        # will not override if the user does provide the scaling factor
        for (t, x, p, j), v in self.flux_mol_phase_comp.items():
            if iscale.get_scaling_factor(v) is None:
                comp = self.config.property_package.get_component(j)
                if comp.is_solvent():
                    if x == 0:
                        prop_feed = self.feed_side.properties_in[t]
                    elif x == 1:
                        prop_feed = self.feed_side.properties_out[t]

                    sf = (iscale.get_scaling_factor(prop_feed.dens_mass_phase['Liq'])
                          / iscale.get_scaling_factor(prop_feed.mw_comp[j])
                          * iscale.get_scaling_factor(prop_feed.pressure)
                          * iscale.get_scaling_factor(self.radius_pore)**2
                          / iscale.get_scaling_factor(prop_feed.visc_d_phase['Liq'])
                          / iscale.get_scaling_factor(self.membrane_thickness_effective))
                    iscale.set_scaling_factor(v, sf)

                if comp.is_solute():
                    #Todo: revisit later
                    # sf = (iscale.get_scaling_factor(self.flux_mol_phase_comp[t, x, 'Liq', 'H2O'])
                    #       / iscale.get_scaling_factor(self.feed_side.properties_in[t].dens_mass_phase['Liq'])
                    #       * iscale.get_scaling_factor(self.feed_side.properties_in[t].mw_comp[j])
                    #       * iscale.get_scaling_factor(self.permeate_side[t, x].conc_mol_phase_comp['Liq', j]))
                    sf = 1e5
                    iscale.set_scaling_factor(v, sf)

        for v in self.rejection_intrinsic_phase_comp.values():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, 1e1)

        for j in self.config.property_package.component_list:
            if j in self.config.property_package.solute_set:
                iscale.set_scaling_factor(self.feed_side.mass_transfer_term[0, 'Liq', j], 1e4)
            else:
                iscale.set_scaling_factor(self.feed_side.mass_transfer_term[0, 'Liq', j], 1)

        for v in self.Kf_comp.values():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, 1e6)

        if self.config.mass_transfer_coefficient == MassTransferCoefficient.spiral_wound:
            for v in self.N_Pe_comp.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)

            for v in self.N_Sc_comp.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-3)

            for v in self.velocity.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e1)

            if iscale.get_scaling_factor(self.channel_height) is None:
                iscale.set_scaling_factor(self.channel_height, 1e3)

        if hasattr(self, 'length'):
            if iscale.get_scaling_factor(self.length) is None:
                iscale.set_scaling_factor(self.length, 1)

        if hasattr(self, 'width'):
            if iscale.get_scaling_factor(self.width) is None:
                iscale.set_scaling_factor(self.width, 1)

        # transforming constraints
        for (t, x, p), con in self.eq_water_flux.items():
            sf = (iscale.get_scaling_factor(self.flux_mol_phase_comp[t, x, p, 'H2O']) \
                 * iscale.get_scaling_factor(self.feed_side.properties_in[t].mw_comp['H2O'])
                 / iscale.get_scaling_factor(self.feed_side.properties_in[t].dens_mass_solvent)
                  )
            iscale.constraint_scaling_transform(con, sf)

        for (t, x, p, j), con in self.eq_solute_solvent_flux.items():
            #todo: revisit sf
            # sf = (iscale.get_constraint_transform_applied_scaling_factor(self.eq_water_flux[t, x, p])
            #       * iscale.get_scaling_factor(self.mixed_permeate[t].conc_mol_phase_comp[p, j]))
            iscale.constraint_scaling_transform(con, 1e2)

        for (t, x, p, j), con in self.eq_solute_flux_concentration_polarization.items():
            #todo: revisit sf
            # sf = (iscale.get_constraint_transform_applied_scaling_factor(self.eq_water_flux[t, x, p])
            #       * iscale.get_scaling_factor(self.mixed_permeate[t].conc_mol_phase_comp[p, j]))
            iscale.constraint_scaling_transform(con, 1e2)

        if self.config.mass_transfer_coefficient == MassTransferCoefficient.spiral_wound:
            for ind, con in self.eq_Kf_comp.items():
                sf = iscale.get_scaling_factor(self.Kf_comp[ind])
                iscale.constraint_scaling_transform(con, sf)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a surface discharge unit
operation.
"""

from pyomo.environ import Constraint, Reference, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_pt, pump_electricity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Travis Arnold"


@declare_process_block_class("SurfaceDischargeZO")
class SurfaceDischargeData(ZeroOrderBaseData):
    """
    Zero-Order model for a surface discharge unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "surface_discharge"
        build_pt(self)
        self._Q = Reference(self.properties[:].flow_vol)

        pump_electricity(self, self._Q)

        self.pipe_distance = Var(self.flowsheet().config.time,
                                 units=pyunits.miles,
                                 doc='Piping distance')

        self.pipe_diameter = Var(self.flowsheet().config.time,
                                 units=pyunits.inches,
                                 doc='Pipe diameter')

        self._fixed_perf_vars.append(self.pipe_distance)
        self._fixed_perf_vars.append(self.pipe_diameter)

        self._perf_var_dict["Pipe Distance"] = self.pipe_distance
        self._perf_var_dict["Pipe Diameter"] = self.pipe_diameter



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a water pumping station.
"""

from pyomo.environ import Reference, units as pyunits, Var
from pyomo.common.config import ConfigValue, In
from idaes.core import declare_process_block_class
from watertap.core import build_pt, ZeroOrderBaseData
from watertap.core.zero_order_electricity import _common
# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("WaterPumpingStationZO")
class WaterPumpingStationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for SW onshore intake operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()
    CONFIG.declare("fix_pump_power", ConfigValue(
        default=True,
        domain=In([True, False]),
        description="Boolean flag for fixing pump power directly.",
        doc="""Indicates whether pump power should be fixed by the user or not.
        **default** - True.
        **Valid values:** {
        **True** -  pump power (variable name "electricity") will be fixed by user and lift_height will not be fixed,
        **False** - pump power (variable name "electricity") is left unfixed and lift_height will be fixed,}"""))

    def build(self):
        super().build()

        self._tech_type = "water_pumping_station"

        build_pt(self)

        # create electricity variable and add to performance dictionary
        _common(self)

        self.lift_height = Var(self.flowsheet().time,
                               initialize=100,
                               units=pyunits.feet,
                               doc="Lift height for pump")
        self.eta_pump = Var(self.flowsheet().time,
                            initialize=0.9,
                            units=pyunits.dimensionless,
                            doc="Efficiency of pump")
        self.eta_motor = Var(self.flowsheet().time,
                             initialize=0.9,
                             units=pyunits.dimensionless,
                             doc="Efficiency of motor")

        self._fixed_perf_vars.append(self.eta_pump)
        self._fixed_perf_vars.append(self.eta_motor)

        if not self.config.fix_pump_power:
            self._fixed_perf_vars.append(self.lift_height)
        else:
            self._fixed_perf_vars.append(self.electricity)

        @self.Constraint(self.flowsheet().time,
                         doc='Constraint for electricity consumption based on '
                         'pump flowrate.')
        def electricity_consumption(b, t):
            A = 3960*pyunits.gallon*pyunits.foot/pyunits.minute/pyunits.horsepower
            return b.electricity[t] == pyunits.convert(
                b.properties[t].flow_vol*b.lift_height[t]/(A*b.eta_pump[t]*b.eta_motor[t]),
                to_units=pyunits.kW)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a fixed bed unit.
operation.
"""

from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_siso, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("FixedBedZO")
class FixedBedZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an Ion exchange unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "fixed_bed"

        build_siso(self)
        constant_intensity(self)
        self.recovery_frac_mass_H2O.fix(1)

        # Chemical demands
        self.acetic_acid_dose = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Dosing rate of acetic acid")
        self.phosphoric_acid_dose = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Dosing rate of phosphoric acid")
        self.ferric_chloride_dose = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Dosing rate of ferric chloride")
        self._fixed_perf_vars.append(self.acetic_acid_dose)
        self._fixed_perf_vars.append(self.phosphoric_acid_dose)
        self._fixed_perf_vars.append(self.ferric_chloride_dose)

        self.acetic_acid_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hr,
            bounds=(0, None),
            doc="Consumption rate of acetic acid")
        self.phosphoric_acid_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hr,
            bounds=(0, None),
            doc="Consumption rate of phosphoric acid")
        self.ferric_chloride_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hr,
            bounds=(0, None),
            doc="Consumption rate of ferric chloride")

        self._perf_var_dict["Acetic Acid Demand"] = self.acetic_acid_demand
        self._perf_var_dict["Phosphoric Acid Demand"] = \
            self.phosphoric_acid_demand
        self._perf_var_dict["Ferric Chlorided Demand"] = \
            self.ferric_chloride_demand

        @self.Constraint(self.flowsheet().time,
                         doc="Acetic acid demand constraint")
        def acetic_acid_demand_equation(b, t):
            return (b.acetic_acid_demand[t] ==
                    pyunits.convert(
                        b.acetic_acid_dose*b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.hr))

        @self.Constraint(self.flowsheet().time,
                         doc="Phosphoric acid demand constraint")
        def phosphoric_acid_demand_equation(b, t):
            return (b.phosphoric_acid_demand[t] ==
                    pyunits.convert(
                        b.phosphoric_acid_dose*b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.hr))

        @self.Constraint(self.flowsheet().time,
                         doc="Acetic acid demand constraint")
        def ferric_chloride_demand_equation(b, t):
            return (b.ferric_chloride_demand[t] ==
                    pyunits.convert(
                        b.ferric_chloride_dose*b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.hr))

        # Activated Carbon demand
        self.activated_carbon_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for activated carbon")
        self.activated_carbon_parameter_a = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for activated carbon demand")
        self.activated_carbon_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for activated carbon demand")
        self._fixed_perf_vars.append(self.activated_carbon_parameter_a)
        self._fixed_perf_vars.append(self.activated_carbon_parameter_b)
        self._perf_var_dict["Activated Carbon Demand"] = \
            self.activated_carbon_demand

        @self.Constraint(self.flowsheet().time,
                         doc="Activated carbon demand constraint")
        def activated_carbon_demand_equation(b, t):
            return (b.activated_carbon_demand[t] ==
                    pyunits.convert(
                        b.activated_carbon_parameter_a *
                        pyunits.convert(b.properties_in[t].flow_vol /
                                        (pyunits.m**3/pyunits.hour),
                                        to_units=pyunits.dimensionless) **
                        b.activated_carbon_parameter_b *
                        b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.hr))

        # Sand demand
        self.sand_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for sand")
        self.sand_parameter_a = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for sand demand")
        self.sand_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for sand demand")
        self._fixed_perf_vars.append(self.sand_parameter_a)
        self._fixed_perf_vars.append(self.sand_parameter_b)
        self._perf_var_dict["Sand Demand"] = \
            self.sand_demand

        @self.Constraint(self.flowsheet().time,
                         doc="Sand demand constraint")
        def sand_demand_equation(b, t):
            return (b.sand_demand[t] ==
                    pyunits.convert(
                        b.sand_parameter_a *
                        pyunits.convert(b.properties_in[t].flow_vol /
                                        (pyunits.m**3/pyunits.hour),
                                        to_units=pyunits.dimensionless) **
                        b.sand_parameter_b *
                        b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.hr))

        # Anthracite demand
        self.anthracite_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for anthracite")
        self.anthracite_parameter_a = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for anthracite demand")
        self.anthracite_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for anthracite demand")
        self._fixed_perf_vars.append(self.anthracite_parameter_a)
        self._fixed_perf_vars.append(self.anthracite_parameter_b)
        self._perf_var_dict["Anthracite Demand"] = \
            self.anthracite_demand

        @self.Constraint(self.flowsheet().time,
                         doc="Anthracite demand constraint")
        def anthracite_demand_equation(b, t):
            return (b.anthracite_demand[t] ==
                    pyunits.convert(
                        b.anthracite_parameter_a *
                        pyunits.convert(b.properties_in[t].flow_vol /
                                        (pyunits.m**3/pyunits.hour),
                                        to_units=pyunits.dimensionless) **
                        b.anthracite_parameter_b *
                        b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.hr))

        # Cationic polymer demand
        self.cationic_polymer_demand = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.hr,
            bounds=(0, None),
            doc="Replacement rate for cationic polymer")
        self.cationic_polymer_parameter_a = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Pre-exponential factor for cationic polymer demand")
        self.cationic_polymer_parameter_b = Var(
            units=pyunits.dimensionless,
            bounds=(None, None),
            doc="Exponential factor for cationic polymer demand")
        self._fixed_perf_vars.append(self.cationic_polymer_parameter_a)
        self._fixed_perf_vars.append(self.cationic_polymer_parameter_b)
        self._perf_var_dict["Cationic Polymer Demand"] = \
            self.cationic_polymer_demand

        @self.Constraint(self.flowsheet().time,
                         doc="Cationic Polymer demand constraint")
        def cationic_polymer_demand_equation(b, t):
            return (b.cationic_polymer_demand[t] ==
                    pyunits.convert(
                        b.cationic_polymer_parameter_a *
                        pyunits.convert(b.properties_in[t].flow_vol /
                                        (pyunits.m**3/pyunits.hour),
                                        to_units=pyunits.dimensionless) **
                        b.cationic_polymer_parameter_b *
                        b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.hr))



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from watertap.tools.parameter_sweep import (LinearSample,
                                            UniformSample,
                                            NormalSample,
                                            LatinHypercubeSample,
                                            _read_output_h5)
import yaml
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

def _yaml_to_dict(yaml_filename):
    """ Reads and stores a yaml file as a dictionary

    Args:
        yaml_filename (str):
            The filename of the yaml file to read.

    Returns:
        input_dict (dict):
            The result of reading the yaml file and translating
            its structure into a dictionary.

    """

    try:
        # Open the yaml file and import the contents into a
        # dictionary with the same structure
        with open(yaml_filename) as fp:
            input_dict = yaml.load(fp, Loader = yaml.FullLoader)

    except:
        raise ValueError('Could not open file %s' % (yaml_filename))

    return input_dict


def get_sweep_params_from_yaml(m, yaml_filename):
    """ Creates a dictionary of swept model parameters specified via yaml file

    This function creates a dictionary of the items to vary during a parameter
    sweep where the variable name, model attribute, and sweeping domain are
    specified in a YAML file.  The YAML file should have the following format::

        A_comp:
            type: NormalSample
            param: fs.RO.A_comp
            mean: 4.0e-12
            std: 0.5e-12

    where the top-level keyword can be any short, easily understood identifier
    for the parameter.  ``type`` must be one of ``LinearSample``, ``UniformSample``,
    ``NormalSample``, or ``LatinHypercubeSample``.  ``param`` must be a valid
    dot-sperated string path to the object attribute (in this case, an RO attribute
    on the flowsheet ``m``) that you wish to vary.  The remaining arguments are
    dependent on the sample type selected.  For ``NormalSample`` information about
    the mean and standard deviation is required.  Consult the ``parameter_sweep``
    help for more information on the different sample classes.

    Args:
        m (pyomo model):
            The flowsheet containing the model to deploy with the parameter sweep
            tool.
        yaml_filename (str):
            The path to the yaml file.

    Returns:
        sweep_params (dict):
            A dictionary containing different instances of parameter sweep samples

    """

    input_dict = _yaml_to_dict(yaml_filename)

    sweep_params = {}

    for param, values in input_dict.items():

        # Find the specified component on the model 
        component = m.find_component(values['param'])

        if component is None:
            raise ValueError(f'Could not acccess attribute {values["param"]}')

        if values['type'] == 'LinearSample':
            sweep_params[param] = LinearSample(component,
                                                values['lower_limit'],
                                                values['upper_limit'],
                                                values['num_samples'])

        elif values['type'] == 'UniformSample':
            sweep_params[param] = UniformSample(component,
                                                values['lower_limit'],
                                                values['upper_limit'])

        elif values['type'] == 'NormalSample':
            sweep_params[param] = NormalSample(component,
                                               values['mean'],
                                               values['std'])

        elif values['type'] == 'LatinHypercubeSample':
            sweep_params[param] = LatinHypercubeSample(component,
                                                       values['lower_limit'],
                                                       values['upper_limit'])

    return sweep_params


def _set_value(component, key, default_value):
   _log.debug(f'Property: {key}')
   _log.debug(f'New Value: {default_value:.6e}, Old Value: {component.value:.6e}')
   component.value = default_value


def set_defaults_from_yaml(m, yaml_filename, verbose=False):
    """ Sets default model values using values stored in a yaml file

    This function reads a yaml file with the structure::

        fs.path.to.attribute_1: 0.123
        fs.path.to.attribute_2: 1.234
        ...

    and uses the (key, default_value) pairs to set default values
    for the attributes in model ``m``.

    Args:
        m (pyomo model):
            The flowsheet containing the model to set default values for
        yaml_filename (str):
            The path to the yaml file.

    Returns:
        N/A

    """

    input_dict = _yaml_to_dict(yaml_filename)

    fail_count = 0

    for key, default_value in input_dict.items():

        # Find the specified component on the model 
        component = m.find_component(key)

        if component is None:
            raise ValueError(f"Could not acccess component {key} on model {m}")

        if component.is_variable_type():
            _set_value(component, key, default_value)
        elif component.is_parameter_type():
            if component.mutable:
                _set_value(component, key, default_value)
            else:
                _log.warn(f"Cannot set value of non-mutable Param {component}")
                fail_count += 1
        else:
            _log.warn(f"Cannot set value of component {component}")
            fail_count += 1

    number_defaults = len(input_dict)
    print(f'Set {number_defaults-fail_count} of {number_defaults} options to default values ({fail_count} failures, see warnings)')
